import {
  useAuth,
  useAuthorizationCache,
  useAppDispatch,
  useSessionCache,
  useQuerybuilderGraph,
  useQuerybuilderHash,
  useML,
  useMLEnabledHash,
  useConfig,
  useQuerybuilderSettings,
  readInSchemaFromBackend,
  assignNewGraphQueryResult,
  useQuerybuilder,
  useVisualizationState,
} from '@graphpolaris/shared/lib/data-access';
import { WebSocketHandler } from '@graphpolaris/shared/lib/data-access/socket';
import Broker from '@graphpolaris/shared/lib/data-access/socket/broker';
import { addInfo } from '@graphpolaris/shared/lib/data-access/store/configSlice';
import { GraphQueryResultFromBackendPayload, queryingBackend } from '@graphpolaris/shared/lib/data-access/store/graphQueryResultSlice';
import { allMLTypes, LinkPredictionInstance, setMLResult } from '@graphpolaris/shared/lib/data-access/store/mlSlice';
import { setQuerybuilderNodes } from '@graphpolaris/shared/lib/data-access/store/querybuilderSlice';
import { SchemaFromBackend } from '@graphpolaris/shared/lib/schema';
import { useEffect } from 'react';
import { SaveStateI, TestDatabaseConnectionResponse, wsGetState, wsGetStates, wsUpdateState, wsSelectState, nilUUID } from './wsState';
import { wsSchemaRequest } from './wsSchema';
import { addSaveState, testedSaveState, selectSaveState, updateSaveStateList, updateSelectedSaveState } from '../store/sessionSlice';
import { URLParams, getParam, deleteParam } from './url';
import { setVisualizationState } from '../store/visualizationSlice';
import { isEqual } from 'lodash-es';

export const EventBus = (props: { onRunQuery: Function; onAuthorized: Function }) => {
  const { login } = useAuth();
  const auth = useAuthorizationCache();
  const dispatch = useAppDispatch();
  const session = useSessionCache();
  const queryHash = useQuerybuilderHash();
  const queryBuilder = useQuerybuilder();
  const mlHash = useMLEnabledHash();
  const visState = useVisualizationState();
  const queryBuilderSettings = useQuerybuilderSettings();

  function loadSaveState(saveStateID: string | undefined, saveStates: Record<string, SaveStateI>) {
    if (saveStateID && saveStates && saveStateID in saveStates) {
      console.debug('Setting state from database', saveStateID, saveStates);
      const state = saveStates[saveStateID];
      if (state) {
        dispatch(setQuerybuilderNodes(state.queryBuilder));
        dispatch(setVisualizationState(state.visualization));
      }
    }
  }

  useEffect(() => {
    Broker.instance().subscribe((data: SchemaFromBackend) => {
      dispatch(readInSchemaFromBackend(data));
      dispatch(addInfo('Schema graph updated'));
    }, 'schema_result');

    Broker.instance().subscribe((data: GraphQueryResultFromBackendPayload) => {
      dispatch(assignNewGraphQueryResult(data));
      dispatch(addInfo('Query Executed!'));
    }, 'query_result');

    // Broker.instance().subscribe((data: QueryBuilderState) => dispatch(setQuerybuilderNodes(data)), 'query_builder_state');

    allMLTypes.forEach((mlType) => {
      Broker.instance().subscribe((data: LinkPredictionInstance[]) => dispatch(setMLResult({ type: mlType, result: data })), mlType);
    });

    Broker.instance().subscribe((data: SaveStateI[]) => {
      dispatch(updateSaveStateList(data));
      console.debug('Save States updated', data);
      const d = Object.fromEntries(data.map((x) => [x.id, x]));
      loadSaveState(session.currentSaveState, d);
      // useEffect(() => {

      // }, [session.currentSaveState, session.saveStates]);
    }, 'save_states');

    Broker.instance().subscribe((data: any) => {}, 'save_state_status');

    Broker.instance().subscribe((data: SaveStateI) => {
      if (data.id !== nilUUID) {
        dispatch(addSaveState(data));
        dispatch(selectSaveState(data.id));
        loadSaveState(data.id, session.saveStates);
      }
    }, 'save_state');

    Broker.instance().subscribe((data: SaveStateI) => {}, 'delete_save_state');

    Broker.instance().subscribe((data: { saveStateID: string; success: boolean }) => {}, 'save_state_selected');

    Broker.instance().subscribe((response: TestDatabaseConnectionResponse) => {
      if (response && response.status === 'success') dispatch(testedSaveState(response.saveStateID));
    }, 'tested_connection');

    login();

    // Setup cleanup
    return () => {
      Broker.instance().unSubscribeAll('schema_result');
      Broker.instance().unSubscribeAll('query_result');

      Broker.instance().unSubscribeAll('save_states');
      Broker.instance().unSubscribeAll('save_state');
      Broker.instance().unSubscribeAll('save_state_status');
      Broker.instance().unSubscribeAll('delete_save_state');
      Broker.instance().unSubscribeAll('tested_connection');
      Broker.instance().unSubscribeAll('save_state_selected');
      // Broker.instance().unSubscribeAll('query_builder_state');
      allMLTypes.forEach((mlType) => {
        Broker.instance().unSubscribeAll(mlType);
      });
    };
  }, []);

  useEffect(() => {
    if (session.currentSaveState) {
      let state = { ...session.saveStates[session.currentSaveState] };
      if (!isEqual(state.queryBuilder, queryBuilder) && state.queryBuilder?.graph?.nodes) {
        console.debug('Updating queryBuilder state', state.queryBuilder, queryBuilder);
        state.queryBuilder = { ...queryBuilder };
        dispatch(updateSelectedSaveState(state));
        wsUpdateState(state);
      }
    }
  }, [queryBuilderSettings, queryHash]);

  useEffect(() => {
    if (session.currentSaveState) {
      let state = { ...session.saveStates[session.currentSaveState] };
      if (!isEqual(state.visualization, visState)) {
        console.debug('Updating visState state', visState);
        state.visualization = { ...visState };
        dispatch(updateSelectedSaveState(state));
        wsUpdateState(state);
      }
    }
  }, [visState]);

  useEffect(() => {
    // New active database
    if (session.currentSaveState && session.currentSaveState !== nilUUID) {
      wsSchemaRequest(session.currentSaveState);
      wsSelectState(session.currentSaveState);
      loadSaveState(session.currentSaveState, session.saveStates);
    }
  }, [session.currentSaveState]);

  useEffect(() => {
    // Newly (un)authorized
    if (auth.authorized && auth.jwt) {
      props.onAuthorized();
      WebSocketHandler.instance()
        .useAuth(auth)
        .connect(() => {
          console.debug('WS connected', session.currentSaveState, window.location.search);

          // Process URL Params
          const paramSaveState = getParam(URLParams.saveState);
          if (paramSaveState && paramSaveState !== nilUUID) {
            wsGetState(paramSaveState);
          } else {
            deleteParam(URLParams.saveState);
          }

          wsGetStates();

          // WebSocketHandler.instance().sendMessage({ //TODO!!
          //   sessionID: auth?.sessionID || '',
          //   key: 'broadcastState',
          //   body: { type: 'subscribe', status: '', value: {} },
          // });
        });
    } else {
      // dispatch(logout());
    }
  }, [auth]);

  useEffect(() => {
    if (!queryBuilder.ignoreReactivity) {
      props.onRunQuery();
      // WebSocketHandler.instance().sendMessage({ //TODO!!
      //   sessionID: auth?.sessionID || '',
      //   key: 'broadcastState',
      //   body: { type: 'query_builder_state', status: '', value: queryBuilder },
      // });
    }
  }, [queryHash, mlHash, queryBuilderSettings]);

  return <div className="hide"></div>;
};
