/**
 * This program has been developed by students from the bachelor Computer Science at
 * Utrecht University within the Software Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */

import { ReceiveMessageI } from '../types';

/**
 * A broker that handles incoming messages from the backend.
 * It works with routingkeys, a listener can subscribe to messages from the backend with a specific routingkey.
 * Possible routingkeys:
 * - query_result:              Contains an object with nodes and edges or a numerical result.
 * - query_translation_result:  Contains the query translated to the database language.
 * - schema_result:             Contains the schema of the users database.
 * - query_status_update:       Contains an update to if a query is being executed.
 * - query_database_error:      Contains the error received from the database.
 * - query_sent:                Contains the message that a query has been send.
 * - gsa_node_result:           Contains a node that has the data for the graph-schema-analytics
 * - gsa_edge_result:           Contains a edge that has the data for the graph-schema-analytics
 */
export default class Broker {
  private static singletonInstance: Broker;
  private listeners: Record<string, Record<string, Function>> = {};
  private catchAllListener: ((data: Record<string, any>, routingKey: string) => void) | undefined;

  /** mostRecentMessages is a dictionary with <routingkey, messageObject>. It stores the most recent message for that routingkey. */
  private mostRecentMessages: Record<string, unknown> = {};

  /** Get the singleton instance of the Broker. */
  public static instance(): Broker {
    if (!this.singletonInstance) this.singletonInstance = new Broker();
    return this.singletonInstance;
  }

  /**
   * Notify all listeners which are subscribed with the specified routingkey.
   * @param {ReceiveMessageI} jsonObject An json object with unknown type.
   * @param {string} routingKey The routing to publish the message to.
   */
  public publish(jsonObject: Record<string, any>, routingKey: string): void {
    this.mostRecentMessages[routingKey] = jsonObject;

    if (this.listeners[routingKey] && Object.keys(this.listeners[routingKey]).length != 0) {
      if (this.catchAllListener) {
        this.catchAllListener(jsonObject, routingKey);
      }
      Object.values(this.listeners[routingKey]).forEach((listener) => listener(jsonObject, routingKey));
      console.debug('%c' + routingKey + ` WS response`, 'background: #222; color: #DBAB2F', jsonObject);
    }
    // If there are no listeners, log the messagep
    else {
      if (this.catchAllListener) {
        this.catchAllListener(jsonObject, routingKey);
        console.debug(routingKey, `catch all used for message with routing key`, jsonObject);
      } else {
        console.debug('%c' + routingKey + ` no listeners for message with routing key`, 'background: #663322; color: #DBAB2F', jsonObject);
      }
    }
  }

  /**
   * Subscribe a listener to messages with the specified routingKey.
   * @param {Function} newListener The listener to subscribe.
   * @param {string} routingKey The routingkey to subscribe to.
   * @param {boolean} consumeMostRecentMessage If true and there is a message for this routingkey available, notify the new listener. Default true.
   */
  public subscribe(
    newListener: Function,
    routingKey: string,
    key: string = (Date.now() + Math.floor(Math.random() * 100)).toString(),
    consumeMostRecentMessage: boolean = false
  ): string {
    if (!this.listeners[routingKey]) this.listeners[routingKey] = {};

    // Don't add a listener twice
    if (!(key in this.listeners[routingKey])) {
      this.listeners[routingKey][key] = newListener;

      // Consume the most recent message
      if (consumeMostRecentMessage && routingKey in this.mostRecentMessages) newListener(this.mostRecentMessages[routingKey], routingKey);
    }

    return key;
  }

  /**
   * Subscribe a listener to messages with the specified routingKey.
   * @param {Function} newListener The listener to subscribe.
   * @param {string} routingKey The routingkey to subscribe to.
   * @param {boolean} consumeMostRecentMessage If true and there is a message for this routingkey available, notify the new listener. Default true.
   */
  public subscribeDefault(newListener: (data: Record<string, any>, routingKey: string) => void): void {
    this.catchAllListener = newListener;
  }

  /**
   * Unsubscribes a listener from messages with specified routingkey.
   * @param {string} routingKey The routing key to unsubscribe from
   * @param {string} listener key of the listener to unsubscribe.
   */
  public unSubscribe(routingKey: string, key: string): void {
    if (this.listeners[routingKey] && key in this.listeners[routingKey]) {
      delete this.listeners[routingKey][key];
    }
  }

  /**
   * Unsubscribes the catch all listener from messages
   */
  public unSubscribeDefault(): void {
    this.catchAllListener = undefined;
  }

  /**
   * Unsubscribes all listeners from messages with specified routingkey.
   * @param {string} routingKey The routing key to unsubscribe from
   */
  public unSubscribeAll(routingKey: string): void {
    this.listeners[routingKey] = {};
  }
}
