import cytoscape from 'cytoscape';
// @ts-ignore
import cise from 'cytoscape-cise';
// @ts-ignore
import coseBilkent from 'cytoscape-cose-bilkent';
// @ts-ignore
import elk from 'cytoscape-elk';

import fcose from 'cytoscape-fcose';
import klay from 'cytoscape-klay';
import dagre from 'cytoscape-dagre';

import Graph from 'graphology';
import { Attributes } from 'graphology-types';
import { Layout } from './layout';
import { ILayoutFactory } from './layout-creator-usecase';
import { CytoscapeLayoutAlgorithms, LayoutAlgorithm } from './types';

cytoscape.use(klay);
cytoscape.use(elk);

export type CytoscapeProvider = 'Cytoscape';

type CytoNode = {
  data: {
    id: string;
    type?: string;
    source?: string;
    target?: string;
    position?: {
      x: number;
      y: number;
    };
    label?: string;
    count?: number;
    color?: string;
  };
};
/**
 * This is the Cytoscape Factory
 */
export class CytoscapeFactory implements ILayoutFactory<CytoscapeLayoutAlgorithms> {
  createLayout(LayoutAlgorithm: CytoscapeLayoutAlgorithms): Cytoscape | null {
    switch (LayoutAlgorithm) {
      case 'Cytoscape_klay':
        //https://github.com/cytoscape/cytoscape.js-klay
        return new CytoscapeKlay();
      case 'Cytoscape_dagre':
        //https://github.com/cytoscape/cytoscape.js-dagre
        return new CytoscapeDagre();
      case 'Cytoscape_elk':
        //https://github.com/cytoscape/cytoscape.js-elk
        return new CytoscapeElk();

      case 'Cytoscape_fcose':
        //https://github.com/iVis-at-Bilkent/cytoscape.js-fcose
        return new CytoscapeFCose();

      case 'Cytoscape_cose-bilkent':
        //https://github.com/cytoscape/cytoscape.js-cose-bilkent
        return new CytoscapeCoseBilkent();

      case 'Cytoscape_cise':
        //https://github.com/iVis-at-Bilkent/cytoscape.js-cise
        return new CytoscapeCise();
      default:
        return null;
    }
  }
}

export abstract class Cytoscape extends Layout<CytoscapeProvider> {
  public verbose = false;

  constructor(public override algorithm: LayoutAlgorithm<CytoscapeProvider>) {
    super('Cytoscape', algorithm);
  }

  // public specialCytoscapeFunction() {
  //   console.log('Only Cytoscape Layouts can do this.');
  // }

  // Takes the schema as input and creates a list of nodes and edges in a format that the layouting algorithm can use.
  public convertToCytoscapeModel(graph: Graph): CytoNode[] {
    const cytonodes: CytoNode[] = [];

    graph.forEachNode((node) => {
      cytonodes.push({
        data: {
          id: node,
          type: 'node',
          label: 'start',
          count: 50,
          color: 'green',
        },
      });
    });

    graph.forEachEdge((edge, _attributes, source, target) => {
      cytonodes.push({
        data: {
          id: edge,
          type: 'edge',
          source: source,
          target: target,
        },
      });
    });

    return cytonodes;
  }

  public setVerbose(verbose: boolean) {
    this.verbose = verbose;
  }

  makeCytoscapeInstance(cytonodes: CytoNode[]) {
    return cytoscape({
      elements: cytonodes,

      headless: true,
      styleEnabled: true,

      style: [
        // the stylesheet for the graph
        {
          selector: 'node',
          style: {
            label: 'data(id)',
            width: '150px',
            height: '50px',
            shape: 'rectangle',
            'background-color': '#fff',
            'text-valign': 'center',
            color: '#333333',
            'border-width': 1,
            'border-color': '#2E1A61',
            'font-size': '10px',
            'padding-left': '30px',
            'padding-right': '30px',
            'padding-top': '50px',
            'padding-bottom': '50px',
          },
          // selector: 'node',
          // style: {
          //   content: 'data(id)',
          //   // label: 'data(id)',
          //   shape: 'round-rectangle',
          //   // width: 'label',
          //   // height: 200,
          //   'font-size': '10px',
          //   'padding-left': '5px',
          //   'padding-right': '5px',
          //   'padding-top': '5px',
          //   'padding-bottom': '5px',
          // },
        },

        {
          selector: 'edge',
          style: {
            'line-color': '#ccc',
            'target-arrow-color': '#ccc',
            'target-arrow-shape': 'triangle',
            'curve-style': 'bezier',
          },
        },
      ],
    });
  }

  updateNodePositions(nodes: cytoscape.NodeCollection, graph: Graph) {
    nodes.forEach((node: cytoscape.NodeSingular) => {
      if (this.verbose) {
        console.log(node.id(), node.position());
      }

      // graph.setNodeAttribute(node.id(), 'x', node.position().x);
      // graph.setNodeAttribute(node.id(), 'y', node.position().y);

      graph.updateNode(node.id(), (attr) => {
        return {
          ...attr,
          x: node.position().x,
          y: node.position().y,
        };
      });

      // console.log(
      //   'width',
      //   node.numericStyle('width'),
      //   'height',
      //   node.numericStyle('height')
      // );
    });
  }
}

/**
 * This is a ConcreteProduct
 */
class CytoscapeKlay extends Cytoscape {
  constructor() {
    super('Cytoscape_klay');
  }

  public override layout(graph: Graph<Attributes, Attributes, Attributes>): void {
    const cytonodes: CytoNode[] = this.convertToCytoscapeModel(graph);

    cytoscape.use(klay);

    const cy = this.makeCytoscapeInstance(cytonodes);

    const layout = cy.layout({
      name: 'klay',
      padding: 40,
      klay: {
        borderSpacing: 40,
        aspectRatio: 0.5,
        compactComponents: true,
      },
      nodeDimensionsIncludeLabels: true,

      // Constrain layout bounds; { x1, y1, x2, y2 } or { x1, y1, w, h }
      // boundingBox: undefined,

      ready: function () {
        // console.info('Layout.ready');
      }, // on layoutready
      stop: function () {
        // console.debug('Layout.stop');
      }, // on layoutstop
    } as any);
    layout.run();

    const nodes = cy.nodes();
    this.updateNodePositions(nodes, graph);

    // var options = {
    //   nodeDimensionsIncludeLabels: false, // Boolean which changes whether label dimensions are included when calculating node dimensions
    //   fit: true, // Whether to fit
    //   padding: 20, // Padding on fit
    //   animate: false, // Whether to transition the node positions
    //   animateFilter: function( node, i ){ return true; }, // Whether to animate specific nodes when animation is on; non-animated nodes immediately go to their final positions
    //   animationDuration: 500, // Duration of animation in ms if enabled
    //   animationEasing: undefined, // Easing of animation if enabled
    //   transform: function( node, pos ){ return pos; }, // A function that applies a transform to the final node position
    //   ready: undefined, // Callback on layoutready
    //   stop: undefined, // Callback on layoutstop
    //   klay: {
    //     // Following descriptions taken from http://layout.rtsys.informatik.uni-kiel.de:9444/Providedlayout.html?algorithm=de.cau.cs.kieler.klay.layered
    //     addUnnecessaryBendpoints: false, // Adds bend points even if an edge does not change direction.
    //     aspectRatio: 1.6, // The aimed aspect ratio of the drawing, that is the quotient of width by height
    //     borderSpacing: 20, // Minimal amount of space to be left to the border
    //     compactComponents: false, // Tries to further compact components (disconnected sub-graphs).
    //     crossingMinimization: 'LAYER_SWEEP', // Strategy for crossing minimization.
    //     /* LAYER_SWEEP The layer sweep algorithm iterates multiple times over the layers, trying to find node orderings that minimize the number of crossings. The algorithm uses randomization to increase the odds of finding a good result. To improve its results, consider increasing the Thoroughness option, which influences the number of iterations done. The Randomization seed also influences results.
    //     INTERACTIVE Orders the nodes of each layer by comparing their positions before the layout algorithm was started. The idea is that the relative order of nodes as it was before layout was applied is not changed. This of course requires valid positions for all nodes to have been set on the input graph before calling the layout algorithm. The interactive layer sweep algorithm uses the Interactive Reference Point option to determine which reference point of nodes are used to compare positions. */
    //     cycleBreaking: 'GREEDY', // Strategy for cycle breaking. Cycle breaking looks for cycles in the graph and determines which edges to reverse to break the cycles. Reversed edges will end up pointing to the opposite direction of regular edges (that is, reversed edges will point left if edges usually point right).
    //     /* GREEDY This algorithm reverses edges greedily. The algorithm tries to avoid edges that have the Priority property set.
    //     INTERACTIVE The interactive algorithm tries to reverse edges that already pointed leftwards in the input graph. This requires node and port coordinates to have been set to sensible values.*/
    //     direction: 'UNDEFINED', // Overall direction of edges: horizontal (right / left) or vertical (down / up)
    //     /* UNDEFINED, RIGHT, LEFT, DOWN, UP */
    //     edgeRouting: 'ORTHOGONAL', // Defines how edges are routed (POLYLINE, ORTHOGONAL, SPLINES)
    //     edgeSpacingFactor: 0.5, // Factor by which the object spacing is multiplied to arrive at the minimal spacing between edges.
    //     feedbackEdges: false, // Whether feedback edges should be highlighted by routing around the nodes.
    //     fixedAlignment: 'NONE', // Tells the BK node placer to use a certain alignment instead of taking the optimal result.  This option should usually be left alone.
    //     /* NONE Chooses the smallest layout from the four possible candidates.
    //     LEFTUP Chooses the left-up candidate from the four possible candidates.
    //     RIGHTUP Chooses the right-up candidate from the four possible candidates.
    //     LEFTDOWN Chooses the left-down candidate from the four possible candidates.
    //     RIGHTDOWN Chooses the right-down candidate from the four possible candidates.
    //     BALANCED Creates a balanced layout from the four possible candidates. */
    //     inLayerSpacingFactor: 1.0, // Factor by which the usual spacing is multiplied to determine the in-layer spacing between objects.
    //     layoutHierarchy: false, // Whether the selected layouter should consider the full hierarchy
    //     linearSegmentsDeflectionDampening: 0.3, // Dampens the movement of nodes to keep the diagram from getting too large.
    //     mergeEdges: false, // Edges that have no ports are merged so they touch the connected nodes at the same points.
    //     mergeHierarchyCrossingEdges: true, // If hierarchical layout is active, hierarchy-crossing edges use as few hierarchical ports as possible.
    //     nodeLayering:'NETWORK_SIMPLEX', // Strategy for node layering.
    //     /* NETWORK_SIMPLEX This algorithm tries to minimize the length of edges. This is the most computationally intensive algorithm. The number of iterations after which it aborts if it hasn't found a result yet can be set with the Maximal Iterations option.
    //     LONGEST_PATH A very simple algorithm that distributes nodes along their longest path to a sink node.
    //     INTERACTIVE Distributes the nodes into layers by comparing their positions before the layout algorithm was started. The idea is that the relative horizontal order of nodes as it was before layout was applied is not changed. This of course requires valid positions for all nodes to have been set on the input graph before calling the layout algorithm. The interactive node layering algorithm uses the Interactive Reference Point option to determine which reference point of nodes are used to compare positions. */
    //     nodePlacement:'BRANDES_KOEPF', // Strategy for Node Placement
    //     /* BRANDES_KOEPF Minimizes the number of edge bends at the expense of diagram size: diagrams drawn with this algorithm are usually higher than diagrams drawn with other algorithms.
    //     LINEAR_SEGMENTS Computes a balanced placement.
    //     INTERACTIVE Tries to keep the preset y coordinates of nodes from the original layout. For dummy nodes, a guess is made to infer their coordinates. Requires the other interactive phase implementations to have run as well.
    //     SIMPLE Minimizes the area at the expense of... well, pretty much everything else. */
    //     randomizationSeed: 1, // Seed used for pseudo-random number generators to control the layout algorithm; 0 means a new seed is generated
    //     routeSelfLoopInside: false, // Whether a self-loop is routed around or inside its node.
    //     separateConnectedComponents: true, // Whether each connected component should be processed separately
    //     spacing: 20, // Overall setting for the minimal amount of space to be left between objects
    //     thoroughness: 7 // How much effort should be spent to produce a nice layout..
    //   },
    //   priority: function( edge ){ return null; }, // Edges with a non-nil value are skipped when greedy edge cycle breaking is enabled
    // };
    // cy.layout( options ).run();
  }
}

function getWidth(node: any) {
  /**
  Calculate the width of a node given its text label `node.data('lbl')`
  */

  // Create element with attributes needed to calculate text size
  const ctx = document.createElement('canvas').getContext('2d');
  const fStyle = node.pstyle('font-style').strValue;
  const size = node.pstyle('font-size').pfValue + 'px';
  const family = node.pstyle('font-family').strValue;
  const weight = node.pstyle('font-weight').strValue;
  ctx!.font = fStyle + ' ' + weight + ' ' + size + ' ' + family;

  // For multiple lines, evaluate the width of the largest line
  const lines = node.data('lbl').split('\n');
  const lengths = lines.map((a: any) => a.length);
  const max_line = lengths.indexOf(Math.max(...lengths));

  // User-defined padding
  const padding = 30;

  return ctx!.measureText(lines[max_line]).width + padding;
}

/**
 * This is a ConcreteProduct
 */
class CytoscapeElk extends Cytoscape {
  constructor() {
    super('Cytoscape_elk');
  }

  public override layout(graph: Graph<Attributes, Attributes, Attributes>, verbose?: boolean): void {
    const cytonodes: CytoNode[] = this.convertToCytoscapeModel(graph);

    const cy = this.makeCytoscapeInstance(cytonodes);

    // const options = {
    //   name: 'elk',
    //   nodeDimensionsIncludeLabels: false, // Boolean which changes whether label dimensions are included when calculating node dimensions
    //   fit: true, // Whether to fit
    //   padding: 20, // Padding on fit
    //   animate: false, // Whether to transition the node positions
    //   animateFilter: function (node: any, i: number) {
    //     return true;
    //   }, // Whether to animate specific nodes when animation is on; non-animated nodes immediately go to their final positions
    //   animationDuration: 500, // Duration of animation in ms if enabled
    //   animationEasing: undefined, // Easing of animation if enabled
    //   transform: function (node: any, pos: Position) {
    //     return pos;
    //   }, // A function that applies a transform to the final node position
    //   ready: undefined, // Callback on layoutready
    //   stop: undefined, // Callback on layoutstop
    //   klay: {
    //     // Following descriptions taken from http://layout.rtsys.informatik.uni-kiel.de:9444/Providedlayout.html?algorithm=de.cau.cs.kieler.klay.layered
    //     addUnnecessaryBendpoints: false, // Adds bend points even if an edge does not change direction.
    //     aspectRatio: 1.6, // The aimed aspect ratio of the drawing, that is the quotient of width by height
    //     borderSpacing: 20, // Minimal amount of space to be left to the border
    //     compactComponents: false, // Tries to further compact components (disconnected sub-graphs).
    //     crossingMinimization: 'LAYER_SWEEP', // Strategy for crossing minimization.
    //     /* LAYER_SWEEP The layer sweep algorithm iterates multiple times over the layers, trying to find node orderings that minimize the number of crossings. The algorithm uses randomization to increase the odds of finding a good result. To improve its results, consider increasing the Thoroughness option, which influences the number of iterations done. The Randomization seed also influences results.
    //     INTERACTIVE Orders the nodes of each layer by comparing their positions before the layout algorithm was started. The idea is that the relative order of nodes as it was before layout was applied is not changed. This of course requires valid positions for all nodes to have been set on the input graph before calling the layout algorithm. The interactive layer sweep algorithm uses the Interactive Reference Point option to determine which reference point of nodes are used to compare positions. */
    //     cycleBreaking: 'GREEDY', // Strategy for cycle breaking. Cycle breaking looks for cycles in the graph and determines which edges to reverse to break the cycles. Reversed edges will end up pointing to the opposite direction of regular edges (that is, reversed edges will point left if edges usually point right).
    //     /* GREEDY This algorithm reverses edges greedily. The algorithm tries to avoid edges that have the Priority property set.
    //     INTERACTIVE The interactive algorithm tries to reverse edges that already pointed leftwards in the input graph. This requires node and port coordinates to have been set to sensible values.*/
    //     direction: 'UNDEFINED', // Overall direction of edges: horizontal (right / left) or vertical (down / up)
    //     /* UNDEFINED, RIGHT, LEFT, DOWN, UP */
    //     edgeRouting: 'ORTHOGONAL', // Defines how edges are routed (POLYLINE, ORTHOGONAL, SPLINES)
    //     edgeSpacingFactor: 0.5, // Factor by which the object spacing is multiplied to arrive at the minimal spacing between edges.
    //     feedbackEdges: false, // Whether feedback edges should be highlighted by routing around the nodes.
    //     fixedAlignment: 'NONE', // Tells the BK node placer to use a certain alignment instead of taking the optimal result.  This option should usually be left alone.
    //     /* NONE Chooses the smallest layout from the four possible candidates.
    //     LEFTUP Chooses the left-up candidate from the four possible candidates.
    //     RIGHTUP Chooses the right-up candidate from the four possible candidates.
    //     LEFTDOWN Chooses the left-down candidate from the four possible candidates.
    //     RIGHTDOWN Chooses the right-down candidate from the four possible candidates.
    //     BALANCED Creates a balanced layout from the four possible candidates. */
    //     inLayerSpacingFactor: 1.0, // Factor by which the usual spacing is multiplied to determine the in-layer spacing between objects.
    //     layoutHierarchy: false, // Whether the selected layouter should consider the full hierarchy
    //     linearSegmentsDeflectionDampening: 0.3, // Dampens the movement of nodes to keep the diagram from getting too large.
    //     mergeEdges: false, // Edges that have no ports are merged so they touch the connected nodes at the same points.
    //     mergeHierarchyCrossingEdges: true, // If hierarchical layout is active, hierarchy-crossing edges use as few hierarchical ports as possible.
    //     nodeLayering: 'NETWORK_SIMPLEX', // Strategy for node layering.
    //     /* NETWORK_SIMPLEX This algorithm tries to minimize the length of edges. This is the most computationally intensive algorithm. The number of iterations after which it aborts if it hasn't found a result yet can be set with the Maximal Iterations option.
    //     LONGEST_PATH A very simple algorithm that distributes nodes along their longest path to a sink node.
    //     INTERACTIVE Distributes the nodes into layers by comparing their positions before the layout algorithm was started. The idea is that the relative horizontal order of nodes as it was before layout was applied is not changed. This of course requires valid positions for all nodes to have been set on the input graph before calling the layout algorithm. The interactive node layering algorithm uses the Interactive Reference Point option to determine which reference point of nodes are used to compare positions. */
    //     nodePlacement: 'BRANDES_KOEPF', // Strategy for Node Placement
    //     /* BRANDES_KOEPF Minimizes the number of edge bends at the expense of diagram size: diagrams drawn with this algorithm are usually higher than diagrams drawn with other algorithms.
    //     LINEAR_SEGMENTS Computes a balanced placement.
    //     INTERACTIVE Tries to keep the preset y coordinates of nodes from the original layout. For dummy nodes, a guess is made to infer their coordinates. Requires the other interactive phase implementations to have run as well.
    //     SIMPLE Minimizes the area at the expense of... well, pretty much everything else. */
    //     randomizationSeed: 1, // Seed used for pseudo-random number generators to control the layout algorithm; 0 means a new seed is generated
    //     routeSelfLoopInside: false, // Whether a self-loop is routed around or inside its node.
    //     separateConnectedComponents: true, // Whether each connected component should be processed separately
    //     spacing: 20, // Overall setting for the minimal amount of space to be left between objects
    //     thoroughness: 7, // How much effort should be spent to produce a nice layout..
    //   },
    //   priority: function (edge: any) {
    //     return null;
    //   }, // Edges with a non-nil value are skipped when greedy edge cycle breaking is enabled
    // };

    const layout = cy.layout({
      name: 'elk',
      fit: true,
      ranker: 'longest-path',
      animate: false,
      elk: {
        zoomToFit: true,
        algorithm: 'mrtree',
        separateConnectedComponents: false,
      },
      ready: function () {
        // console.log('Layout.start');
      }, // on layoutready
      stop: function () {
        // console.log('Layout.stop');
      }, // on layoutstop
    } as any);
    layout.run();

    this.updateNodePositions(cy.nodes(), graph);
  }
}

/**
 * This is a ConcreteProduct
 */
class CytoscapeDagre extends Cytoscape {
  constructor() {
    super('Cytoscape_dagre');
    cytoscape.use(dagre);
  }

  public override layout(graph: Graph<Attributes, Attributes, Attributes>, verbose?: boolean): void {
    const cytonodes: CytoNode[] = this.convertToCytoscapeModel(graph);

    const cy = this.makeCytoscapeInstance(cytonodes);

    const layout = cy.layout({
      name: 'dagre',
      animate: false,
      // acyclicer: 'greedy',
      ranker: 'longest-path',
      spacingFactor: 0.7,
      ready: function () {
        // console.log('Layout.start');
      }, // on layoutready
      stop: function () {
        // // console.log('Layout.stop');
      }, // on layoutstop
    } as any);
    layout.run();

    this.updateNodePositions(cy.nodes(), graph);
  }
}

/**
 * This is a ConcreteProduct
 */
class CytoscapeFCose extends Cytoscape {
  constructor() {
    super('Cytoscape_fcose');
    cytoscape.use(fcose);
  }

  public override layout(graph: Graph<Attributes, Attributes, Attributes>, verbose?: boolean): void {
    const cytonodes: CytoNode[] = this.convertToCytoscapeModel(graph);

    const cy = this.makeCytoscapeInstance(cytonodes);

    const layout = cy.layout({
      name: 'fcose',
      animate: false,
      ready: function () {
        // // console.log('Layout.start');
      }, // on layoutready
      stop: function () {
        // // console.log('Layout.stop');
      }, // on layoutstop
    } as any);
    layout.run();

    const nodes = cy.nodes();
    this.updateNodePositions(nodes, graph);
  }
}

/**
 * This is a ConcreteProduct
 */
class CytoscapeCoseBilkent extends Cytoscape {
  constructor() {
    super('Cytoscape_cose-bilkent');
    cytoscape.use(coseBilkent);
  }

  public override layout(graph: Graph<Attributes, Attributes, Attributes>, verbose?: boolean): void {
    const cytonodes: CytoNode[] = this.convertToCytoscapeModel(graph);

    const cy = this.makeCytoscapeInstance(cytonodes);

    const layout = cy.layout({
      name: 'cose-bilkent',
      animate: false,
      ready: function () {
        // console.log('Layout.start');
      }, // on layoutready
      stop: function () {
        // console.log('Layout.stop');
      }, // on layoutstop
    } as any);
    layout.run();

    this.updateNodePositions(cy.nodes(), graph);
  }
}

/**
 * This is a ConcreteProduct
 */
class CytoscapeCise extends Cytoscape {
  constructor() {
    super('Cytoscape_cise');
    cytoscape.use(cise);
  }

  public override layout(graph: Graph<Attributes, Attributes, Attributes>, verbose?: boolean): void {
    const cytonodes: CytoNode[] = this.convertToCytoscapeModel(graph);

    const cy = this.makeCytoscapeInstance(cytonodes);

    const layout = cy.layout({
      name: 'cise',
      ready: function () {
        // console.log('Layout.start');
      }, // on layoutready
      stop: function () {
        // console.log('Layout.stop');
      }, // on layoutstop
    } as any);
    layout.run();

    this.updateNodePositions(cy.nodes(), graph);
  }
}
